#!/usr/bin/perl
#!/usr/local/bin/perl
# ==============================================================================================================================
# FileName      : exportToRad
# Subjet        : export CSE objects in Proviso format to RAD objects.
# Exit codes    : 0 = Success
#                 1 = Error
#                 2 = Warning
# ==============================================================================================================================
# Netcool/Proviso                                        Copyright(c) MICROMUSE
# ==============================================================================================================================

# ==============================================================================================================================
# Requirements
# ==============================================================================================================================
#require 5.8.0;

# ==============================================================================================================================
# Pragmas
# ==============================================================================================================================
use strict;
use sigtrap;

# ==============================================================================================================================
# Standard PERL modules
# ==============================================================================================================================
use POSIX;
use Getopt::Long;
use File::Basename;
use File::Spec;
use File::Copy;

# ==============================================================================================================================
# Library path
# ==============================================================================================================================

# ==============================================================================================================================
# Specific PERL modules
# ==============================================================================================================================

# ==============================================================================================================================
# Trap signals
# ==============================================================================================================================
#$SIG{ 'INT' } = $SIG{ 'QUIT' } = $SIG{ 'TERM' } = 'IGNORE';

# ==============================================================================================================================
# Global variables
# ==============================================================================================================================
my $VERSION							= "1.09";
my $PROGNAME						= basename($0,".pl") || basename($0);
my $ERROR_OK						= 0;
my $ERROR_KO						= 1;
my $EXIT_STATUS_OK					= 0;
my $EXIT_STATUS_ERROR				= 1;
my $EXIT_STATUS_WARNING				= 2;
my $DEV_NULL						= File::Spec->devnull();
my $TEMPLATES_FILE_NAME				= File::Spec->catfile("conf", "radTemplates.conf");
my $TEMPLATES_BCK_FILE_NAME				= File::Spec->catfile("conf", "radTemplates.bck");
my $INST_TEMPLATES_BCK_FILE_NAME			= File::Spec->catfile("conf", "radTemplatesLastSyncInst.bck");
my $CSE_BCK_FILE_NAME				= File::Spec->catfile("conf", "radCse.bck");
my $SE_BCK_FILE_NAME				       = File::Spec->catfile("conf", "radSe.bck");
my $NPATH_BCK_FILE_NAME				= File::Spec->catfile("conf", "radNPath.bck");
my $USERS_BCK_FILE_NAME				= File::Spec->catfile("conf", "radUsers.bck");

# Actions
my $CREATE_TEMPLATES				= 1;
my $DELETE_TEMPLATES				= 2;
my $CREATE_INSTANCES				= 3;
my $DELETE_INSTANCES				= 4;
my $CLEAN_BACKUP				= 5;

# Arguments
my $debug							= 0;
my $action							= $CREATE_INSTANCES;
my $nbWarnings						= 0;
my @templatesFilter					= ();
my $outputFile;
my $templatesFile;
my $emitPermissions      = 0;  # Cahalanb 20/11/09
                             # New paramater permission
                             # If set, do not output permission statements to the Radhsell command file
my $suppressBlank					= 0;

my $propertyGroupPermission			= "GroupForInstancePermissions";
my $propertyProvisoResourceID		= "ProvisoResourceID";
my $propertyProvisoGroupID			= "ProvisoGroupID";

my $templatesBckFile;
my $lastSyncInstTemplBckFile;
my $cseBckFile;
my $npathBckFile;
my $usersBckFile;
my $templatesFilterResmgr;
# Cahalanb 31/03/10 PMR79264 - holder for the SE file
my $seBckFile;
# PMR79264

my %newTemplates					= ();		# List of service templates read recently
my %oldTemplates					= ();		# List of service templates read previously
my %newInstances					= ();		# List of service instances extracted recently from Proviso
my %oldInstances					= ();		# List of service instances extracted previously
#my %newIndexMappings   					= ();
#my %oldIndexMappings   					= ();
my %SEHash 						= ();
my %newSubElements				= ();
my %oldSubElements				= ();

my $dateTime;  # Cahalanb 24/11/09 - declare variable to hold the date and time.

# ==============================================================================================================================
# Usage description
# ==============================================================================================================================
my $SHORT_USAGE = "$PROGNAME.pl [-help|-?] [OPTIONS]";

# Cahalanb 20/11/09  Modify Usage string to display Help on our new Parameter noPermissions
my $USAGE					= "
NAME
	$PROGNAME.pl - generate RAD shell functions to create service templates and service
	instances for Netcool/RAD database from CSE objects in Proviso.

SYNOPSIS
	$SHORT_USAGE

DESCRIPTION
	The utility writes on the standard output, RAD shell functions to create service templates
	or to create service instances for Netcool/RAD database from Composite Sub-Elements in
	Proviso. By default only service instances are written.

OPTIONS
	The following options are supported:

	-help, -?		Show this help and exit.
	-version		Display version information and exit.
	-debug			Print  diagnostic information.
	-createTemplates	Generate RAD shell functions to create service templates.
	-createInstances	Generate RAD shell functions to create service instances (by default).
	-deleteTemplates	Generate RAD shell functions to delete service templates.
	-deleteInstances	Generate RAD shell functions to delete service instances.
	-templatesFile file	A path of the file that describes service templates to create in
				the Netcool/RAD GUI (\$PVMHOME/conf/radTemplates.conf by default).
	-templates name,name...	List of service templates name used to limit the number of service
				instances to create.
	-output outputFile	The results are written in the \"outputFile\" file.
	-clean			Delete backup files.
 	-emitPermissions  	emit RAD shell functions that grant permissions to Users.
 	-suppressBlank		Suppress propagating Proviso properties that have blank value

FILE
	The templates file describes service templates to create in the Netcool/RAD GUI.
	A complete description consists of the following:

	- A set of service templates

		<Template icon=\"icon1.gif\">template1</Template>
		<Template icon=\"icon2.gif\">template2</Template>

		This <Template> tag specifies the name of a template to create and the name of the file
		for the icon which represents the service template in the Netcool/RAD GUI.

	- An optional set of properties

		<Property>ProvisoReportID</Property>
		<Property>ProvisoSiteID</Property>

		This <Property> tag specifies the name of a property to associate with each service template.

	- An optional set of rad shell functions to associate with each service template

		<CommonShell>setTemplateDescription(\"<TEMPLATE_NAME>\", \"Service Status\");</CommonShell>
		<CommonShell>createSlaForTemplate(\"<TEMPLATE_NAME>\", \"slaName\");</CommonShell>

		This <CommonShell> tag describes a RAD shell function to generate with each service template.
		The variable <TEMPLATE_NAME> is replaced by the name of the template by the converToRad utility.

	- An optional set of rad shell functions to add

		<SingleShell>addNumericFunctionsPolicy(\"Example1\");</SingleShell>

		This <SingleShell> tag describes a RAD shell function to generate.

	- An optional filter of CSE path

		<PathFilter>~NOC Reporting*</PathFilter>

		This <PathFilter> tag describes the npath to use to filter CSE objects.

	These tags should be contained within a <Templates> tag.

EXIT STATUS
	The following exit values are returned:

	0		All CSE objects are exported successfully.

	1		An error occurred.

	2		A warning occured during the conversion.";

# ==============================================================================================================================
# Command line options
# ==============================================================================================================================
my %KEYS_USAGES = (
  'help|?'				=> sub {print "$USAGE\n"; exit $EXIT_STATUS_OK;},
  'version'				=> sub {print "version: $VERSION\n"; exit $EXIT_STATUS_OK;},
  'debug!'				=> \$debug,
  'createTemplates'		=> sub {$action = $CREATE_TEMPLATES;},
  'createInstances'		=> sub {$action = $CREATE_INSTANCES;},
  'deleteTemplates'		=> sub {$action = $DELETE_TEMPLATES;},
  'deleteInstances'		=> sub {$action = $DELETE_INSTANCES;},
  'clean'				=> sub {$action = $CLEAN_BACKUP;},
  'templatesFile=s'		=> \$templatesFile,
  'templates=s'			=> \@templatesFilter,
  'output=s'			=> \$outputFile,
  'emitPermissions'   	=> \$emitPermissions,  # Cahalanb 20/11/09  variable for the new option
  'suppressBlank'		=> \$suppressBlank
);

# ==============================================================================================================================
# displayError
# ------------------------------------------------------------------------------------------------------------------------------
# Description  :
# Input        : error message
# Return       : $ERROR_KO
# ==============================================================================================================================
sub displayError
{
	my($errorMsg) = @_;
	print STDERR "$errorMsg\n";
	return $ERROR_KO;
}

# ==============================================================================================================================
# createError
# ------------------------------------------------------------------------------------------------------------------------------
# Description  :
# Input        : error message
# Return       : $ERROR_KO
# ==============================================================================================================================
sub createError
{
	my($errorMsg) = @_;
	print STDERR "$PROGNAME\tError\t$errorMsg.\n";
	return $ERROR_KO;
}

# ==============================================================================================================================
# createWarning
# ------------------------------------------------------------------------------------------------------------------------------
# Description  :
# Input        : warning message
# Return       : none
# ==============================================================================================================================
sub createWarning
{
	my($warningMsg) = @_;
	print STDERR "$PROGNAME\tWarning\t$warningMsg.\n";
	$nbWarnings += 1;
}

# ==============================================================================================================================
# createDebug
# ------------------------------------------------------------------------------------------------------------------------------
# Description  :
# Input        : debug message
# Return       :
# ==============================================================================================================================
sub createDebug
{
	my($debugMsg) = @_;
	if ($debug)
	{
		print STDERR "$PROGNAME\tDebug\t$debugMsg.\n";
	}
}

# ==============================================================================================================================
# getHeaderParent
# ------------------------------------------------------------------------------------------------------------------------------
# Description  :
# Input        : a service instance
# Return       : the service instance header parent or undef
# ==============================================================================================================================
sub getHeaderParent
{
	my($refInstances, $instance)	= @_;
	my $result						= undef;

	if (exists $refInstances->{$instance})
	{
		if (not exists $refInstances->{$instance}{parent})
		{
			$result = $instance;
		}
		else
		{
			$result = getHeaderParent($refInstances, $refInstances->{$instance}{parent});
		}
	}
	return $result;
}

# Cahalanb 31/03/10 PMR79264 - exportSE - generate a file with SE records from the current system
# ==============================================================================================================================
# exportSE
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : export SE objects from Proviso to a file
# Input        :
# Return       :
# ==============================================================================================================================
sub exportSE
{
	$dateTime = localtime();
	createDebug("Beginning exportSE at $dateTime");
	
	my($destFile)			= @_;
	my $err					= $ERROR_OK;

	# Build Proviso command to get CSE objects from Proviso
	my $cmd = 'resmgr -export se -colNames "dbIndex label tag';
	if (exists $newTemplates{properties})
	{
		foreach my $property (sort keys %{$newTemplates{properties}})
		{
			$cmd .= " seprp.$property:value";
		}
	}

	# Option -ListForced
	if (exists $newTemplates{properties})
	{
		my $flFirst = 1;
		# cahalanb additional " inserted in front of  -ListForced
		$cmd .= '" -ListForced "';
		foreach my $property (sort keys %{$newTemplates{properties}})
		{
			if ($flFirst == 1)
			{
				$cmd .= "seprp.$property:value \\\"\\\"";
				$flFirst = 0;
			}
			else
			{
				$cmd .= " seprp.$property:value \\\"\\\"";
			}
		}
		$cmd .= '"';
	}
	
	# Option -filter
	$cmd .= " -filter \"tag(in$templatesFilterResmgr)\"";

	# Option -file
  	$cmd .= " -file $destFile";
  
	$cmd .= " 2>$DEV_NULL " unless ($debug);
	createDebug("$cmd");
	
	#$cmd = "cp export.se.dat.org $destFile";
	#createDebug("Modified Command '$cmd'");

	# Execute the export
	!system "$cmd" or ($err = createError("Cannot run command \'$cmd\'"));

	$dateTime = localtime();
	createDebug("Finishing exportSE at $dateTime");

	return $err;
}
# Cahalanb 31/03/10 PMR79264 

# ==============================================================================================================================
# exportCSE
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : export CSE objects from Proviso to a file
# Input        :
# Return       :
# ==============================================================================================================================
sub exportCSE
{
	$dateTime = localtime();
	createDebug("Beginning exportCSE at $dateTime");

	my($destFile)			= @_;
	my $err					= $ERROR_OK;

	# Build Proviso command to get CSE objects from Proviso
	my $cmd = 'resmgr -export se -colNames "dbIndex label tag';
	if (exists $newTemplates{properties})
	{
		foreach my $property (sort keys %{$newTemplates{properties}})
		{
			$cmd .= " seprp.$property:value";
		}
	}
	$cmd .= ' cse.name cse.coIndex"';

	# Option -ListForced
	if (exists $newTemplates{properties})
	{
		my $flFirst = 1;
		$cmd .= ' -ListForced "';
		foreach my $property (sort keys %{$newTemplates{properties}})
		{
			if ($flFirst == 1)
			{
				$cmd .= "seprp.$property:value \\\"\\\"";
				$flFirst = 0;
			}
			else
			{
				$cmd .= " seprp.$property:value \\\"\\\"";
			}
		}
		$cmd .= '"';
	}
	# Option -filter
	$cmd .= " -filter \"tag(in$templatesFilterResmgr)\"";

	# Option -file
	$cmd .= " -file $destFile";

	$cmd .= " 2>$DEV_NULL " unless ($debug);
	createDebug("$cmd");

	#$cmd = "cp export.cse.dat.org $destFile";
	#createDebug("Modified Command: '$cmd'");
	
	# Execute the export
	!system "$cmd" or ($err = createError("Cannot run command \'$cmd\'"));

	$dateTime = localtime();
	createDebug("Finishing exportCSE at $dateTime");

	return $err;
}

# ==============================================================================================================================
# buildCSE
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : build CSE objects from a file
# Input        :
# Return       :
# ==============================================================================================================================
sub buildCSE
{
	$dateTime = localtime();
	createDebug("Beginning buildCSE at $dateTime");

	my($refTemplates, $refInstances, $path)	= @_;
	my $err									= $ERROR_OK;
	my $isOpen								= 0;

	open fdCSE, "<$path" or ($err = createError("Cannot open file \'$path\'"));

	# Read CSE
	if ($err == $ERROR_OK)
	{
		$isOpen = 1;
		while (<fdCSE>)
		{
			# Remove LF and jump comments
			chomp;
			next if (/^(#.*|\s*)$/);
			s/"/ /g;

			
			# Get dbIndex (service instance name), display name and template name
			my @fields = split('\|_\|');
			my $dbIndex = shift(@fields);

			if (not defined($dbIndex))
			{
				createWarning("Field \'dbIndex\' unexpected in line \'$_\'");
				next;
			}
			

			my $label = shift(@fields);
			my $tag = shift(@fields);
			if (not defined($tag))
			{
				createWarning("Field \'tag\' unexpected in line \'$_\'");
				next;
			}
			$refInstances->{$dbIndex}{"templateName"} = $tag;
			$refInstances->{$dbIndex}{"displayName"} = $label;
			$refInstances->{$dbIndex}{"action"} = "delete";

			# Get users preferences for this service template
			if (exists $refTemplates->{properties})
			{
				foreach my $property (sort keys %{$refTemplates->{properties}})
				{
					my $valueProp = shift(@fields);
					unless ( $suppressBlank && ($valueProp eq "") )
					{
						$refInstances->{$dbIndex}{"properties"}{$property} = $valueProp;
					}
				}
			}
			# Get children service instances
			if (scalar(@fields) % 2)
			{
				createWarning("Number of fields unexpected in line \'$_\'");
				next;
			}
			
			
			my $cseName = shift(@fields);
			while (defined($cseName))
			{
				my $cseCoIndex = shift(@fields);
					
				if (not defined($cseCoIndex))
				{
					createWarning("Field \'cse.coIndex\' unexpected in line \'$_\'");
					last;
				}
				# Add children
				$refInstances->{$dbIndex}{"children"}{$cseCoIndex} = undef;

				$cseName = shift(@fields);
			}
		}
	}

	if ($isOpen == 1)
	{
		close fdCSE;
	}

	
	$dateTime = localtime();
	createDebug("Finishing buildCSE at $dateTime");

	return $err;
}

# ==============================================================================================================================
# buildParentChildRelations
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : link the children to the parents, prune missing children
# Input        :
# Return       :
# ==============================================================================================================================
sub buildParentChildRelations
{
	$dateTime = localtime();
	createDebug("Beginning buildParentChildRelations at $dateTime");

	my($refInstances)	= @_;

	# Build dependencies parent-child between CSE
	foreach my $instance (keys %$refInstances)
	{
		# If this CSE has children
		if (exists $refInstances->{$instance}{children})
		{
			my @children = keys %{$refInstances->{$instance}{children}};
			foreach my $child (@children)
			{
				if (exists $refInstances->{$child})
				{
					# And if a child exists among CSE, set its parent with this CSE
					$refInstances->{$child}{"parent"} = $instance;
				}
				else
				{
					# Else if a child doesn't exist, it is removed of children list
					delete $refInstances->{$instance}{children}{$child};
				}
			}
			# The list of children is empty, the list is destroyed
			unless (scalar(keys %{$refInstances->{$instance}{children}}))
			{
				delete $refInstances->{$instance}{children};
			}
		}
	}

}

# ==============================================================================================================================
# exportNPath
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : export npath associated with CSE objects from Proviso to a file
# Input        :
# Return       :
# ==============================================================================================================================
sub exportNPath
{
	$dateTime = localtime();
	createDebug("Beginning exportNPath at $dateTime");

	my($destFile)			= @_;
	my $err					= $ERROR_OK;

	# Build Proviso command to get npath of CSE objects from Proviso
	my $cmd = "resmgr -export segp -colNames \"dbIndex npath se.dbIndex\" -filter \"se.tag(in$templatesFilterResmgr)";
	if (defined $newTemplates{pathFilter})
	{
		$cmd .= " npath($newTemplates{pathFilter})";
	}
	$cmd .= '"';

	# Option -file
	$cmd .= " -file $destFile";

	$cmd .= " 2>$DEV_NULL " unless ($debug);
	createDebug("$cmd");
	
	#$cmd = "cp export.npath.dat.org $destFile";
	#createDebug("Modified Command: '$cmd'");

	# Execute the export
	!system "$cmd" or ($err = createError("Cannot run command \'$cmd\'"));

	$dateTime = localtime();
	createDebug("Finishing exportNPath at $dateTime");

	return $err;
}

# ==============================================================================================================================
# addNPath
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : add npath from a file to CSE objects
# Input        :
# Return       :
# ==============================================================================================================================
sub addNPath
{
	my($refInstances, $path)	= @_;
	my $err						= $ERROR_OK;
	my $isOpen					= 0;

	open fdNPath, "<$path" or ($err = createError("Cannot open file \'$path\'"));

	# Read npath
	if ($err == $ERROR_OK)
	{
		$isOpen = 1;
		while (<fdNPath>)
		{
			# Remove LF and jump comments
			chomp;
			next if (/^(#.*|\s*)$/);
			s/"/ /g;

			# Get npath and service instance name
			my($dbIndex, $npath, $seDbIndex) = (split('\|_\|'))[0, 1, 2];
			if (not defined($npath))
			{
				createWarning("Field \'npath\' unexpected in line \'$_\'");
				next;
			}
			if (not defined($seDbIndex))
			{
				createWarning("Field \'se.dbIndex\' unexpected in line \'$_\'");
				next;
			}
			if ((exists $refInstances->{$seDbIndex}) && (not exists $refInstances->{$seDbIndex}{parent}))
			{
				$refInstances->{$seDbIndex}{"npath"} = $npath;
				$refInstances->{$seDbIndex}{"groupID"} = $dbIndex;
			}
		}
	}
	if ($isOpen == 1)
	{
		close fdNPath;
	}
	return $err;
}

# ==============================================================================================================================
# exportUsers
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : export users associated with CSE objects from Proviso to a file
# Input        :
# Return       :
# ==============================================================================================================================
sub exportUsers
{
	$dateTime = localtime();
	createDebug("Beginning exportUsers at $dateTime");

	my($destFile)			= @_;
	my $err					= $ERROR_OK;

	# Build Proviso command to get npath of user objects from Proviso
	my $cmd = 'resmgr -export segp -colNames "npath rusr.name" -filter "rusr.type(PROVISO USER)"';

	# Option -file
	$cmd .= " -file $destFile";

	$cmd .= " 2>$DEV_NULL " unless ($debug);
	createDebug("$cmd");

	 #$cmd = "cp export.ruser.dat.org $destFile";
	 #createDebug("Modified Command: '$cmd'");

	# Execute the export
	!system "$cmd" or ($err = createError("Cannot run command \'$cmd\'"));

	$dateTime = localtime();
	createDebug("Finishing exportUsers at $dateTime");

	return $err;
}

# ==============================================================================================================================
# addUsers
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : add users from a file to CSE objects
# Input        :
# Return       :
# ==============================================================================================================================
sub addUsers
{
	my($refInstances, $path)	= @_;
	my $err						= $ERROR_OK;
	my $isOpen					= 0;

	open fdUsers, "$path" or ($err = createError("Cannot open file \'$path\'"));

	# Read users
	if ($err == $ERROR_OK)
	{
		$isOpen = 1;
		while (<fdUsers>)
		{
			# Remove LF and jump comments
			chomp;
			next if (/^(#.*|\s*)$/);
			s/"/ /g;

			# Get npath and user name
			my($npath, $userName) = (split('\|_\|'))[0, 1];
			if (not defined($npath))
			{
				createWarning("Field \'npath\' unexpected in line \'$_\'");
				next;
			}
			if (not defined($userName))
			{
				createWarning("Field \'ruser.name\' unexpected in line \'$_\'");
				next;
			}
			unless ($npath =~ /.*\~$/)
			{
				$npath .= '~';
			}
			if ($userName =~ /^SSSW\\\\(.*)$/)
			{
				$userName = $1;
			}
			foreach my $instance (keys %$refInstances)
			{
				if ((not exists $refInstances->{$instance}{parent}) && (exists $refInstances->{$instance}{npath}))
				{
					my $npathInstance = $refInstances->{$instance}{npath};
					unless ($npathInstance =~ /.*\~$/)
					{
						$npathInstance .= '~';
					}
					# If the user is associated with this service instance
					if (($npathInstance =~ /^$npath/) || ($npathInstance eq $npath))
					{
						$refInstances->{$instance}{"users"}{$userName} = undef;
					}
				}
			}
		}
	}
	if ($isOpen == 1)
	{
		close fdUsers;
	}
	return $err;
}

# ==============================================================================================================================
# getTemplates
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : get Service Templates from description file
# Input        :
# Return       :
# ==============================================================================================================================
sub getTemplates
{
	$dateTime = localtime();
	createDebug("Beginning getTemplatesat $dateTime");

	my($refTemplates, $path)	= @_;
	my $err						= $ERROR_OK;
	my $isOpen					= 0;

	createDebug("getTemplates path=$path");

	if ($err == $ERROR_OK)
	{
		open TEMPLATES_FD, "<$path" or ($err = createError("Cannot open service templates file \'$path\'"));
	}
	if ($err == $ERROR_OK)
	{
		$isOpen = 1;
		while (<TEMPLATES_FD>)
		{
			# Remove LF and jump comments
			chomp;
			next if (/^(\s*<!--.*-->\s*|\s*)$/);

			if (/<(Property)>(.*)<\/\1>/)
			{
				# Process a property
				$_ = $2;
				s/\s*//g;
				unless (/^(ProvisoResourceID|ProvisoGroupID|)$/i)
				{
					$refTemplates->{"properties"}{$_}{"action"} = "delete";
				}
			}
			elsif (/<(GroupPermissionProperty)>(.*)<\/\1>/)
			{
				$_ = $2;
				s/\s*//g;
				$refTemplates->{groupPermissionProperty} = $_;
				unless ( exists $refTemplates->{"properties"}{$_} )
				{
					$refTemplates->{"properties"}{$_}{"action"} = "delete";
				}
			}
			elsif (/<(GroupPermissionPropertyInvalidChars)>(.*)<\/\1>/)
			{
				$_ = $2;
				s/\s*//g;
				$refTemplates->{groupPermissionPropertyInvalidChars} = $_;
			}
			elsif (/<(PathFilter)>(.*)<\/\1>/)
			{
				# Process the path filter
				$refTemplates->{"pathFilter"} = $2;
			}
			elsif (/<(Template) icon=(["'])(.*)\2>(.*)<\/\1>/)
			{
				# Process a service template
				my $name = $4;
				my $icon = $3;
				$_ = $name;
				s/\s*//g;
				$name = $_;
				$_ = $icon;
				s/\s*//g;
				$icon = $_;
				$refTemplates->{"templates"}{$name}{"icon"} = $icon;
				$refTemplates->{"templates"}{$name}{"action"} = "delete";
			}
			elsif (/<(CommonShell)>(.*)<\/\1>/)
			{
				# Process a common radshell function for each service template
				$refTemplates->{"commonShell"}{$2}{"action"} = "delete";
			}
			elsif (/<(SingleShell)>(.*)<\/\1>/)
			{
				# Process a single radshell function for all service templates
				$refTemplates->{"singleShell"}{$2}{"action"} = "delete";
			}
			elsif (/^\s*(<Templates>)|(<\/Templates>)\s*$/)
			{
				# NOP
			}
			else
			{
				$err = createError("Unexpected format in the service templates file \'$path\' in line $_");
				last;
			}
		}
	}
	if ($err == $ERROR_OK)
	{
		unless (exists $refTemplates->{templates})
		{
			$err = createError("Service templates not defined");
		}
	}
	if ($isOpen == 1)
	{
		close TEMPLATES_FD;
	}

	$dateTime = localtime();
	createDebug("Finishing getTemplatesat at $dateTime");

	return $err;
}

# ==============================================================================================================================
# generateRadShellTemplates
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : generate the rad shell functions to create Service Templates
# Input        :
# Return       :
# ==============================================================================================================================
sub generateRadShellTemplates
{

	$dateTime = localtime();
	createDebug("Beginning generateRadShellTemplates at $dateTime");

	my $err					= $ERROR_OK;

	# Check if a property or a rad function needs to be added
	if (exists $newTemplates{properties})
	{
		foreach my $property (keys %{$newTemplates{properties}})
		{
			unless ((exists $oldTemplates{properties}) && (exists $oldTemplates{properties}{$property})) 
			{
				$newTemplates{properties}{$property}{"action"} = "create";
			}
		}
	}
	if (exists $newTemplates{commonShell})
	{
		foreach my $shell (keys %{$newTemplates{commonShell}})
		{
			unless ((exists $oldTemplates{commonShell}) && (exists $oldTemplates{commonShell}{$shell})) 
			{
				$newTemplates{commonShell}{$shell}{"action"} = "create";
			}
		}
	}
	# Add and/or delete if necessary service templates
	if (exists $newTemplates{templates})
	{
		foreach my $template (sort keys %{$newTemplates{templates}})
		{
			$newTemplates{templates}{$template}{"action"} = "none";
			if (exists $oldTemplates{templates}{$template})
			{
				# This template exists already
				$oldTemplates{templates}{$template}{"action"} = "none";
				if ($newTemplates{templates}{$template}{icon} ne $oldTemplates{templates}{$template}{icon})
				{
					# But it must be updated
					$newTemplates{templates}{$template}{"action"} = "update";
					print "deleteTemplate(\"$template\");\n";
				}
			}
			else
			{
				# This template is new
				$newTemplates{templates}{$template}{"action"} = "create";
			}
			if (($newTemplates{templates}{$template}{action} eq "update") ||
				($newTemplates{templates}{$template}{action} eq "create"))
			{
				# Add properties and rad functions
				print "createTemplate(\"$template\", \"$newTemplates{templates}{$template}{icon}\");\n";
				print "addUserPreferencesForTemplate(\"$template\", \"ProvisoResourceID\", \"0\");\n";
				print "addUserPreferencesForTemplate(\"$template\", \"ProvisoGroupID\", \"0\");\n";
				if (exists $newTemplates{properties})
				{
					foreach my $property (sort keys %{$newTemplates{properties}})
					{
						print "addUserPreferencesForTemplate(\"$template\", \"$property\", \"0\");\n";
					}
				}
				if (exists $newTemplates{commonShell})
				{
					foreach my $shell (keys %{$newTemplates{commonShell}})
					{
						$_ = $shell;
						s/<TEMPLATE_NAME>/$template/g;
						print "$_\n";
					}
				}
			}
			else
			{
				# Add properties if necessary
				if (exists $newTemplates{properties})
				{
					foreach my $property (sort keys %{$newTemplates{properties}})
					{
						if ($newTemplates{properties}{$property}{action} eq "create") 
						{
							print "addUserPreferencesForTemplate(\"$template\", \"$property\", \"0\");\n";
						}
					}
				}
				# Add rad functions if necessary
				if (exists $newTemplates{commonShell})
				{
					foreach my $shell (keys %{$newTemplates{commonShell}})
					{
						if ($newTemplates{commonShell}{$shell}{"action"} eq "create")
						{
							$_ = $shell;
							s/<TEMPLATE_NAME>/$template/g;
							print "$_\n";
						}
					}
				}
			}
		}
	}
	# Add single shells
	if (exists $newTemplates{singleShell})
	{
		foreach my $shell (keys %{$newTemplates{singleShell}})
		{
			unless ((exists $oldTemplates{singleShell}) && (exists $oldTemplates{singleShell}{$shell})) 
			{
				print "$shell\n";
			}
		}
	}
	# Delete useless service templates
	if (exists $oldTemplates{templates})
	{
		foreach my $template (sort keys %{$oldTemplates{templates}})
		{
			if ($oldTemplates{templates}{$template}{action} eq "delete")
			{
				print "deleteTemplate(\"$template\");\n";
			}
		}
	}

	$dateTime = localtime();
	createDebug("Finishing generateRadShellTemplates at $dateTime");

	return $err;
}

# ==============================================================================================================================
# generateTemplates
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : generate the RAD shell command to create Service Templates
# Input        :
# Return       :
# ==============================================================================================================================
sub generateTemplates
{
	$dateTime = localtime();
	createDebug("Beginning generateTemplates at $dateTime");

	my $err					= $ERROR_OK;

	createDebug("generateTemplates");

	# Read the service templates file
	if ($err == $ERROR_OK)
	{
		$err = getTemplates(\%newTemplates, $templatesFile);
	}
	# Get templates from backup file
	if ($err == $ERROR_OK)
	{
		$err = getTemplates(\%oldTemplates, $templatesBckFile) if -r $templatesBckFile;
	}
	# Generate the rad shell for the creation of templates
	if ($err == $ERROR_OK)
	{
		$err = generateRadShellTemplates();
	}
	# Backup templates description file
	if ($err == $ERROR_OK)
	{
		copy($templatesFile, $templatesBckFile) or ($err = createError("Cannot copy \'$templatesFile\' to \'$templatesBckFile\'"));
	}

	$dateTime = localtime();
	createDebug("Finishing generateTemplates at $dateTime");

	return $err;
}

# ==============================================================================================================================
# removeBackupTemplates
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : remove the backup file of service templates
# Input        :
# Return       :
# ==============================================================================================================================
sub removeBackupTemplates
{
	my $err		= $ERROR_OK;

	if (-e $templatesBckFile)
	{
		unless (unlink $templatesBckFile)
		{
			$err = createError("Cannot delete backup file \'$templatesBckFile\'");
		}
	}
	return $err;
}
# ==============================================================================================================================
# deleteTemplates
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : generate the RAD shell command to delete Service Templates
# Input        :
# Return       :
# ==============================================================================================================================
sub deleteTemplates
{
	my $err					= $ERROR_OK;

	createDebug("deleteTemplates");

	# Get templates from backup file
	if ($err == $ERROR_OK)
	{
		$err = getTemplates(\%oldTemplates, $templatesBckFile) if -r $templatesBckFile;
	}
	# Generate the rad shell for the destruction of templates
	if ($err == $ERROR_OK)
	{
		$err = generateRadShellTemplates();
	}
	# Remove the backup file
	if ($err == $ERROR_OK)
	{
		$err = removeBackupTemplates();
	}
	return $err;
}

# ==============================================================================================================================
# getInstancesFromProviso
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : get Service Instances from Proviso
# Input        :
# Return       :
# ==============================================================================================================================
sub getInstancesFromProviso
{
	$dateTime = localtime();
	createDebug("Beginning getInstancesFromProviso at $dateTime");

	my $err					= $ERROR_OK;
	my $cseBckFileTmp		= $cseBckFile . ".tmp";
	my $npathBckFileTmp		= $npathBckFile . ".tmp";
	my $usersBckFileTmp		= $usersBckFile . ".tmp";
	my $seBckFileTmp		= $seBckFile . ".tmp";
	my $flInit				= 0;

	# Read the service templates file
	if ($err == $ERROR_OK)
	{
		$err = getTemplates(\%newTemplates, $templatesFile);
	}
	# Create a filter for Proviso command from templates file if the filter isn't passed in argument
	if ($err == $ERROR_OK)
	{
		if (not defined($templatesFilterResmgr))
		{
			foreach my $template (sort keys %{$newTemplates{templates}})
			{
				if ($flInit == 0)
				{
					$templatesFilterResmgr .= "\'$template\'";
					$flInit = 1;
				}
				else
				{
					$templatesFilterResmgr .= ",\'$template\'";
				}
			}
		}
	}
	# Get CSE objects from Proviso
	if ($err == $ERROR_OK)
	{
		$err = exportCSE($cseBckFileTmp);
	}
	if ($err == $ERROR_OK)
	{
		$err = exportSE($seBckFileTmp);
	}
	if ($err == $ERROR_OK)
	{		
		$err = buildCSE(\%newTemplates, \%newInstances, $cseBckFileTmp);
	}
	if ($err == $ERROR_OK)
	{		
		$err = buildSE(\%newTemplates, $seBckFileTmp, \%newSubElements);
	}
	if ($err == $ERROR_OK)
	{		
		$err = addLeafSE(\%newInstances, \%newSubElements);
	}
	if ($err == $ERROR_OK)
	{		
		$err = buildParentChildRelations(\%newInstances);
	}
	# Add npath
	if ($err == $ERROR_OK)
	{
		$err = exportNPath($npathBckFileTmp);
	}
	if ($err == $ERROR_OK)
	{
		$err = addNPath(\%newInstances, $npathBckFileTmp);
	}
	# Add users
	if ($err == $ERROR_OK)
	{
		$err = exportUsers($usersBckFileTmp);
	}
	if ($err == $ERROR_OK)
	{
		$err = addUsers(\%newInstances, $usersBckFileTmp);
	}

	$dateTime = localtime();
	createDebug("Finishing getInstancesFromProviso at $dateTime");

	return $err;
}

# ==============================================================================================================================
# backupInstances
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : backup Service Instances from Proviso and the current templates file
# Input        :
# Return       :
# ==============================================================================================================================
sub backupInstances
{
	my $err		= $ERROR_OK;
	my @files	= ($cseBckFile, $npathBckFile, $usersBckFile, $seBckFile);

	foreach my $file (@files)
	{
		my $tmpFile = $file . ".tmp";
		unless (rename $tmpFile, $file)
		{
			$err = createError("Cannot rename \'$tmpFile\' to \'$file\'");
			last;
		}
	}
	if ($err == $ERROR_OK)
	{
		copy($templatesFile, $lastSyncInstTemplBckFile) or ($err = createError("Cannot copy \'$templatesFile\' to \'$lastSyncInstTemplBckFile\'"));
	}
	return $err;
}

# ==============================================================================================================================
# getInstancesFromRestore
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : get Service Instances from Backup files
# Input        :
# Return       :
# ==============================================================================================================================
sub getInstancesFromRestore
{
	my $err					= $ERROR_OK;

	# Read the service templates file
	if ($err == $ERROR_OK)
	{
		$err = getTemplates(\%oldTemplates, $lastSyncInstTemplBckFile) if -r $lastSyncInstTemplBckFile;
	}
	# Get CSE objects from Backup
	if ($err == $ERROR_OK)
	{
		$err = buildCSE(\%oldTemplates, \%oldInstances, $cseBckFile) if -r $cseBckFile;
	}
	if ($err == $ERROR_OK)
	{
		$err = buildSE(\%oldTemplates, $seBckFile, \%oldSubElements) if -r $seBckFile;
	}
	if ($err == $ERROR_OK)
	{
		$err = addLeafSE(\%oldInstances, \%oldSubElements);
	}
	if ($err == $ERROR_OK)
	{
		$err = buildParentChildRelations(\%oldInstances);
	}
 	# Add npath
	if ($err == $ERROR_OK)
	{
		$err = addNPath(\%oldInstances, $npathBckFile) if -r $npathBckFile;
	}
	# Add users
	if ($err == $ERROR_OK)
	{
		$err = addUsers(\%oldInstances, $usersBckFile) if -r $usersBckFile;
	}
	return $err;
}

# ==============================================================================================================================
# removeBackupInstances
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : remove backup files of service instances
# Input        :
# Return       :
# ==============================================================================================================================
sub removeBackupInstances
{
	my $err		= $ERROR_OK;
	my @files	= ($cseBckFile, $npathBckFile, $usersBckFile, $lastSyncInstTemplBckFile, $seBckFile);

	foreach my $file (@files)
	{
		if (-e $file)
		{
			unless (unlink $file)
			{
				$err = createError("Cannot delete backup file \'$file\'");
				last;
			}
		}
	}
	return $err;
}

# ==============================================================================================================================
# generateRadShellInstances
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : generate the rad shell functions to create Service Instances
# Input        :
# Return       :
# ==============================================================================================================================
sub generateRadShellInstances
{
	$dateTime = localtime();
	createDebug("Beginning generateRadShellInstances at $dateTime");

	my $err					= $ERROR_OK;

	# Add, update or delete if necessary service instances
	foreach my $instance (sort keys %newInstances)
	{
		$newInstances{$instance}{"action"} = "none";
		if (exists $oldInstances{$instance})
		{
			# This instance exists already
			$oldInstances{$instance}{"action"} = "none";

			if ($newInstances{$instance}{templateName} ne $oldInstances{$instance}{templateName})
			{
				$newInstances{$instance}{"action"} = "update";
				print "updateServiceInstanceTemplate(\"$instance\", \"$newInstances{$instance}{templateName}\");\n";
			}
			if ($newInstances{$instance}{displayName} ne $oldInstances{$instance}{displayName})
			{
				$newInstances{$instance}{"action"} = "update";
				print "updateServiceInstanceDisplayName(\"$instance\", \"$newInstances{$instance}{displayName}\");\n";
			}
		}
		else
		{
			# This instance is new
			$newInstances{$instance}{"action"} = "create";

			print "addServiceInstance(new String[] {\"$newInstances{$instance}{templateName}\"}, \"$instance\", " .
				"\"$newInstances{$instance}{displayName}\", \"\", \"Standard\", null, null);\n";
		}
	}
	# Delete useless service instances
	foreach my $instance (sort keys %oldInstances)
	{
		if ($oldInstances{$instance}{action} eq "delete")
		{
			if (exists $oldInstances{$instance}{users})
			{
				foreach my $user (sort keys %{$oldInstances{$instance}{users}})
				{
					print "revokeObjectPrivilege(USER, \"$user\", INSTANCE, \"$instance\", \"RAD_VIEW_INSTANCE\");\n";
					print "revokeObjectPrivilege(USER, \"$user\", INSTANCE, \"$instance\", \"RAD_USER\");\n";
				}
			}
			print "deleteServiceInstance(new String[] {\"$instance\"});\n";
		}
	}

	$dateTime = localtime();
	createDebug("Finishing generateRadShellInstances at $dateTime");

	return $err;
}

# ==============================================================================================================================
# getGroupPermissionValue
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : 
# Input        : 
# Return       : 
# ==============================================================================================================================
sub getGroupPermissionValue
{
	my ($instance, $refTemplates, $refInstances)	= @_;
	my $groupPermissionValue				= undef;

	if (exists $refTemplates->{groupPermissionProperty})
	{
		my $groupPermissionProperty = $refTemplates->{groupPermissionProperty};

		# check for group permission property on this service instance
		if ( (exists $refInstances->{$instance}{properties}) && 
		     (exists $refInstances->{$instance}{properties}{$groupPermissionProperty}) &&
		     ($refInstances->{$instance}{properties}{$groupPermissionProperty} ne "") 
		     )
		{
			$groupPermissionValue = $refInstances->{$instance}{properties}{$groupPermissionProperty};
		}
		else
		{
			# check for group permission property on top-most parent instance
			my $refHeader = getHeaderParent($refInstances, $instance);
			if (defined $refHeader)
			{
				if ( (exists $refInstances->{$refHeader}{properties}) && 
				     (exists $refInstances->{$refHeader}{properties}{$groupPermissionProperty}) &&
				     ($refInstances->{$refHeader}{properties}{$groupPermissionProperty} ne "") 
				     )
				{
					$groupPermissionValue = $refInstances->{$refHeader}{properties}{$groupPermissionProperty};
				}
			}
		}
	}

	if (defined $groupPermissionValue)
	{
		if (exists $refTemplates->{groupPermissionPropertyInvalidChars})
		{
			my $groupPermissionPropertyInvalidChars = $refTemplates->{groupPermissionPropertyInvalidChars};
			$groupPermissionValue =~ s/[$groupPermissionPropertyInvalidChars]//g;
		}
		createDebug("getGroupPermissionValue: instance $instance = $groupPermissionValue");
	}
	else
	{
		createDebug("getGroupPermissionValue: instance $instance = <undefind>");
	}

	return $groupPermissionValue;
}

# ==============================================================================================================================
# generateRadShellUserPreferences
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : generate the rad shell functions to create User Preferences
# Input        :
# Return       :
# ==============================================================================================================================
sub generateRadShellUserPreferences
{
	$dateTime = localtime();
	createDebug("Beginning generateRadShellUserPreferences at $dateTime");

	my $err					= $ERROR_OK;

	# For each service instance, add or update a fieldvalue pair (user preference) for the instance
	foreach my $instance (sort keys %newInstances)
	{
		if ($newInstances{$instance}{action} eq "create")
		{
			# This instance is just created, all User Preferences must be created

			# ProvisoResourceID User Preference
			print "addUserPreferencesForInstance(\"$instance\", \"ProvisoResourceID\", \"$instance\");\n";

			# ProvisoGroupID User Preference
			my $headerInstance = getHeaderParent(\%newInstances, $instance);
			if ((defined $headerInstance) && (exists $newInstances{$headerInstance}{groupID}))
			{
				print "addUserPreferencesForInstance(\"$instance\", \"ProvisoGroupID\", " .
						"\"$newInstances{$headerInstance}{groupID}\");\n";
			}
			# Other properties User Preference
			if (exists $newInstances{$instance}{properties})
			{
				foreach my $property (sort keys %{$newInstances{$instance}{properties}})
				{
					print "addUserPreferencesForInstance(\"$instance\", \"$property\", " .
							"\"$newInstances{$instance}{properties}{$property}\");\n";
				}
			}
			my $groupPermissionValue = getGroupPermissionValue($instance, \%newTemplates, \%newInstances);
			if (defined $groupPermissionValue)
			{
				print "addUserPreferencesForInstance(\"$instance\", \"$propertyGroupPermission\", " .
						"\"$groupPermissionValue\");\n";
			}
		}
		else
		{
			# This instance exists already, check if User Preferences must be created or updated
			# ProvisoResourceID User Preference changes never

			# ProvisoGroupID User Preference
			my $newHeader = getHeaderParent(\%newInstances, $instance);
			if ((defined $newHeader) && (exists $newInstances{$newHeader}{groupID}))
			{
				my $oldHeader = getHeaderParent(\%oldInstances, $instance);
				if ((defined $oldHeader) && (exists $oldInstances{$oldHeader}{groupID}))
				{
					if ($newInstances{$newHeader}{groupID} ne $oldInstances{$oldHeader}{groupID})
					{
						print "updateUserPreferencesForInstance(\"$instance\", \"ProvisoGroupID\", " .
								"\"$newInstances{$newHeader}{groupID}\");\n";
					}
				}
				else
				{
					print "addUserPreferencesForInstance(\"$instance\", \"ProvisoGroupID\", " .
							"\"$newInstances{$newHeader}{groupID}\");\n";
				}
			}
			# Other properties User Preference
			if (exists $newInstances{$instance}{properties})
			{
				foreach my $property (sort keys %{$newInstances{$instance}{properties}})
				{
					if ((exists $oldInstances{$instance}{properties}) && (exists $oldInstances{$instance}{properties}{$property}))
					{
						if ($newInstances{$instance}{properties}{$property} ne $oldInstances{$instance}{properties}{$property})
						{
							print "updateUserPreferencesForInstance(\"$instance\", \"$property\", " .
									"\"$newInstances{$instance}{properties}{$property}\");\n";
						}
					}
					else
					{
						print "addUserPreferencesForInstance(\"$instance\", \"$property\", " .
								"\"$newInstances{$instance}{properties}{$property}\");\n";
					}
				}
			}
			# group permission property
			my $groupPermissionValueNew = getGroupPermissionValue($instance, \%newTemplates, \%newInstances);
			my $groupPermissionValueOld = getGroupPermissionValue($instance, \%oldTemplates, \%oldInstances);
			if (defined $groupPermissionValueNew)
			{
				if (defined $groupPermissionValueOld)
				{
					if ($groupPermissionValueNew ne $groupPermissionValueOld)
					{
						print "updateUserPreferencesForInstance(\"$instance\", \"$propertyGroupPermission\", " .
							"\"$groupPermissionValueNew\");\n";
					}
				}
				else
				{
					print "addUserPreferencesForInstance(\"$instance\", \"$propertyGroupPermission\", " .
						"\"$groupPermissionValueNew\");\n";
				}
			}
		}
	}

	$dateTime = localtime();
	createDebug("Finishing generateRadShellUserPreferences at $dateTime");

	return $err;
}

# ==============================================================================================================================
# generateRadShellDependencies
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : generate the RAD shell command to create Service Instances Dependencies
# Input        :
# Return       :
# ==============================================================================================================================
sub generateRadShellDependencies
{
	$dateTime = localtime();
	createDebug("Beginning generateRadShellDependencies at $dateTime");

	my $err					= $ERROR_OK;

	# For each service instance, add or delete dependencies between instances
	# If a service instance disappeared, its dependencies are destroyed automatically 
	foreach my $instance (sort keys %newInstances)
	{
		if (exists $newInstances{$instance}{children})
		{
			foreach my $child (sort keys %{$newInstances{$instance}{children}})
			{
				if ($newInstances{$instance}{action} eq "create")
				{
					# This instance is just created, all its dependencies must be created
					print "addServiceInstanceDependency(\"$instance\", \"$child\");\n";
				}
				else
				{
					# This instance exists already
					unless ((exists $oldInstances{$instance}{children}) && (exists $oldInstances{$instance}{children}{$child}))
					{
						# And it has a new child
						print "addServiceInstanceDependency(\"$instance\", \"$child\");\n";
					}
				}
			}
		}
		if ((exists $oldInstances{$instance}) && (exists $oldInstances{$instance}{children}))
		{
			foreach my $child (sort keys %{$oldInstances{$instance}{children}})
			{
				unless ((exists $newInstances{$instance}{children}) && (exists $newInstances{$instance}{children}{$child}))
				{
					# A dependency disappeared
					if (exists $newInstances{$child})
					{
						# But the child still exists
						print "deleteServiceInstanceDependency(\"$instance\", \"$child\");\n";
					}
				}
			}
		}
	}

	$dateTime = localtime();
	createDebug("Finishing generateRadShellDependencies at $dateTime");

	return $err;
}

# ==============================================================================================================================
# generateRadShellObjectPrivileges
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : generate the RAD shell command to create Objects Privileges
# Input        :
# Return       :
# ==============================================================================================================================
sub generateRadShellObjectPrivileges
{
	$dateTime = localtime();
	createDebug("Beginning generateRadShellObjectPrivileges at $dateTime");

	my $err					= $ERROR_OK;

	# Firstly, for each service instance, revoke an useless instance privilege
	# If a service instance doesn't exist anymore, its privileges are already revoked
	foreach my $instance (sort keys %oldInstances)
	{
		if ($oldInstances{$instance}{action} ne "delete")
		{
			if (exists $oldInstances{$instance}{users})
			{
				foreach my $user (sort keys %{$oldInstances{$instance}{users}})
				{
					unless ((exists $newInstances{$instance}{users}) && (exists $newInstances{$instance}{users}{$user}))
					{
						print "revokeObjectPrivilege(USER, \"$user\", INSTANCE, \"$instance\", \"RAD_VIEW_INSTANCE\");\n";
						print "revokeObjectPrivilege(USER, \"$user\", INSTANCE, \"$instance\", \"RAD_USER\");\n";
					}
				}
			}
		}
	}
	# For each service instance, grant an instance privilege (such as RAD_VIEW_INSTANCE)
	# to an instance as well as all its children recursively.  This is useful
	# when you want a user/group to be able to see all down the hierarchy
	# without having to go to each instance in the GUI
	foreach my $instance (sort keys %newInstances)
	{
		if (exists $newInstances{$instance}{users})
		{
			foreach my $user (sort keys %{$newInstances{$instance}{users}})
			{
				unless ((exists $oldInstances{$instance}) && (exists $oldInstances{$instance}{users})
					&& (exists $oldInstances{$instance}{users}{$user}))
				{
					# If this instance is new or its user is new, the user must be associated to the instance
					print "grantObjectPrivilege(USER, \"$user\", INSTANCE, \"$instance\", \"RAD_USER\");\n";
					print "grantObjectPrivilege(USER, \"$user\", INSTANCE, \"$instance\", \"RAD_VIEW_INSTANCE\");\n";
					print "grantPrivilegeToAllChildren(USER, \"$user\", \"$instance\", \"RAD_USER\");\n";
					print "grantPrivilegeToAllChildren(USER, \"$user\", \"$instance\", \"RAD_VIEW_INSTANCE\");\n";
				}
			}
		}
	}

	$dateTime = localtime();
	createDebug("Finishing generateRadShellObjectPrivileges at $dateTime");

	return $err;
}

# ==============================================================================================================================
# generateInstances
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : generate the RAD shell command to create Service Instances
# Input        :
# Return       :
# ==============================================================================================================================
sub generateInstances
{
	$dateTime = localtime();
	createDebug("Beginning generateInstances at $dateTime");

	my $err					= $ERROR_OK;

	createDebug("generateInstances");

	# Get service instances from Proviso
	if ($err == $ERROR_OK)
	{
		$err = getInstancesFromProviso();
	}
	# Get service instances from Backup files
	if ($err == $ERROR_OK)
	{
		$err = getInstancesFromRestore();
	}
	# Generate the rad shell for the creation of instances
	if ($err == $ERROR_OK)
	{
		$err = generateRadShellInstances();
	}
	# Generate the rad shell for the creation of properties
	if ($err == $ERROR_OK)
	{
		$err = generateRadShellUserPreferences();
	}
	# Generate the rad shell for the creation of dependencies
	if ($err == $ERROR_OK)
	{
		$err = generateRadShellDependencies();
	}
	# Generate the rad shell for the creation of object privileges
	if ($err == $ERROR_OK)
	{
    	if($emitPermissions) # Cahalanb 20/11/09 - Check if the permissions flag has been set. If so, do not output Grant permission commands.
    	{
      		$err = generateRadShellObjectPrivileges();
    	}
	}
	# Backup services instances
	if ($err == $ERROR_OK)
	{
		$err = backupInstances();
	}

	$dateTime = localtime();
	createDebug("Finishing generateInstances at $dateTime");

	return $err;
}

# ==============================================================================================================================
# deleteInstances
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : generate the RAD shell command to delete Service Instances
# Input        :
# Return       :
# ==============================================================================================================================
sub deleteInstances
{
	my $err					= $ERROR_OK;

	createDebug("deleteInstances");

	# Get service instances from Backup files
	if ($err == $ERROR_OK)
	{
		$err = getInstancesFromRestore();
	}
	# Generate the rad shell for the destruction of instances
	# The destruction of properties, dependencies and object privileges is useless
	if ($err == $ERROR_OK)
	{
		$err = generateRadShellInstances();
	}
	# Remove the backup files
	if ($err == $ERROR_OK)
	{
		$err = removeBackupInstances();
	}
	return $err;
}

# ==============================================================================================================================
# buildSE
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : build SE objects from a file
# Input        :
# Return       :
# ==============================================================================================================================
sub buildSE
{
	$dateTime = localtime();
	createDebug("Beginning buildSE at $dateTime");

	my($refTemplates, $path, $refSubElements)	= @_;
	my $err									= $ERROR_OK;
	my $isOpen								= 0;

	open fdSE, "<$path" or ($err = createError("Cannot open file \'$path\'"));

	# Read SE
	if ($err == $ERROR_OK)
	{
		$isOpen = 1;
		while (<fdSE>)
		{
			# Remove LF and jump comments
			chomp;
			next if (/^(#.*|\s*)$/);
			s/"/ /g;

			
			# Get dbIndex (service instance name), display name and template name
			my @fields = split('\|_\|');
			my $dbIndex = shift(@fields);

			if (not defined($dbIndex))
			{
				createWarning("Field \'dbIndex\' unexpected in line \'$_\'");
				next;
			}
			
			my $label = shift(@fields);
			my $tag = shift(@fields);
			if (not defined($tag))
			{
				createWarning("Field \'tag\' unexpected in line \'$_\'");
				next;
			}
			$refSubElements->{$dbIndex}{"templateName"} = $tag;
			$refSubElements->{$dbIndex}{"displayName"} = $label;
			$refSubElements->{$dbIndex}{"action"} = "delete";

			# Get users preferences for this service template
			if (exists $refTemplates->{properties})
			{
				foreach my $property (sort keys %{$refTemplates->{properties}})
				{
					my $valueProp = shift(@fields);
					unless ( $suppressBlank && ($valueProp eq "") )
					{
						$refSubElements->{$dbIndex}{"properties"}{$property} = $valueProp;					
					}
				}
			}
		}
	}

	if ($isOpen == 1)
	{
		close fdSE;
	}
	
	$dateTime = localtime();
	createDebug("Finishing buildSE at $dateTime");

	return $err;
}

# ==============================================================================================================================
# addLeafSE
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : copy in leaf-level CSE Model sub-elements
# Input        : $refInstances - Hash model representing all CSEs
#              : $refSubElements - hash with all sub-elements
# Return       : Success / Failure
# ==============================================================================================================================
sub addLeafSE
{
	my ($refInstances, $refSubElements) = @_;
	my $err		= $ERROR_OK;
	
	foreach my $instance (sort keys %{$refInstances})
	{
		if (exists $refInstances->{$instance}{children})
		{
			my @children = keys %{$refInstances->{$instance}{children}};
			foreach my $child (@children)
			{
				if ( (not exists $refInstances->{$child}) && (exists $refSubElements->{$child}) )
				{
					createDebug("addLeafSE: instance $instance adding child $child");
					$refInstances->{$child}{"templateName"} = $refSubElements->{$child}{"templateName"};
					$refInstances->{$child}{"displayName"} = $refSubElements->{$child}{"displayName"};
					$refInstances->{$child}{"action"} = $refSubElements->{$child}{"action"};
					if (exists $refSubElements->{$child}{"properties"})
					{
						$refInstances->{$child}{"properties"} = $refSubElements->{$child}{"properties"};
					}
					$refInstances->{$child}{"children"} = ();
				}
			}
		}
	}
}

# ==============================================================================================================================
# initialize
# ------------------------------------------------------------------------------------------------------------------------------
# Description  : initialize environnement
# Input        :
# Return       :
# ==============================================================================================================================
sub initialize
{
	my $err					= $ERROR_OK;
	my $flInit				= 0;
	my %filter				= ();

	# Parse and check command line arguments
	unless (GetOptions(%KEYS_USAGES))
	{
		$err = displayError("$SHORT_USAGE");
	}
	# Check Proviso path
	if ($err == $ERROR_OK)
	{
		if (not defined($ENV{'PVMHOME'}))
		{
			$err = createError("PVMHOME variable is not defined in environment");
		}
	}
	if ($err == $ERROR_OK)
	{
		if (not -d $ENV{'PVMHOME'})
		{
			$err = createError("The path \'$ENV{'PVMHOME'}\' in PVMHOME variable is not a directory");
		}
	}
	# Set paths
	if ($err == $ERROR_OK)
	{
		$templatesBckFile = File::Spec->catfile($ENV{'PVMHOME'}, $TEMPLATES_BCK_FILE_NAME);
		$lastSyncInstTemplBckFile = File::Spec->catfile($ENV{'PVMHOME'}, $INST_TEMPLATES_BCK_FILE_NAME);
		$cseBckFile = File::Spec->catfile($ENV{'PVMHOME'}, $CSE_BCK_FILE_NAME);
		$seBckFile = File::Spec->catfile($ENV{'PVMHOME'}, $SE_BCK_FILE_NAME);
		$npathBckFile = File::Spec->catfile($ENV{'PVMHOME'}, $NPATH_BCK_FILE_NAME);
		$usersBckFile = File::Spec->catfile($ENV{'PVMHOME'}, $USERS_BCK_FILE_NAME);

		# Set the default templates file if necessary
		if (not defined($templatesFile))
		{
			$templatesFile = File::Spec->catfile($ENV{'PVMHOME'}, $TEMPLATES_FILE_NAME);
		}
	}
	# Open the output file
	if ($err == $ERROR_OK)
	{
		if (defined($outputFile))
		{
			open STDOUT, ">$outputFile" or ($err = createError("Cannot open output file $outputFile"));
		}
	}
	# Create a filter for Proviso command
	if ($err == $ERROR_OK)
	{
		@templatesFilter = split /,/, join(',', @templatesFilter);
		foreach my $template (@templatesFilter)
		{
			$filter{$template} = undef;
		}
		foreach my $template (sort keys %filter)
		{
			if ($flInit == 0)
			{
				$templatesFilterResmgr .= "\'$template\'";
				$flInit = 1;
			}
			else
			{
				$templatesFilterResmgr .= ",\'$template\'";
			}
		}
	}
	return $err;
}

# ==============================================================================================================================
# main
# ==============================================================================================================================
sub main
{
	$dateTime = localtime();
	createDebug("Beginning Main at $dateTime");

	my $err = $ERROR_OK;

	# Initialization
	if ($err == $ERROR_OK)
	{
		$err = initialize();
	}
	if ($err == $ERROR_OK)
	{
		if ($action == $CREATE_TEMPLATES)
		{
			# Generate service templates
			$err = generateTemplates();
		}
		elsif ($action == $CREATE_INSTANCES)
		{

			# Generate service instances
			$err = generateInstances();
		}
		elsif ($action == $DELETE_INSTANCES)
		{
			# Delete service instances
			$err = deleteInstances();
		}
		elsif ($action == $DELETE_TEMPLATES)
		{
			# Delete service templates
			$err = deleteTemplates();
		}
		elsif ($action == $CLEAN_BACKUP)
		{
			# Delete backup files
			$err = removeBackupInstances();
			if ($err == $ERROR_OK)
			{
				$err = removeBackupTemplates();
			}
		}
	}

	$dateTime = localtime();
	createDebug("Finishing Main at $dateTime");

	return $err;
}
# ==============================================================================================================================
my $err = $ERROR_OK;
my $status = $EXIT_STATUS_OK;

# Main
$err = main();

# Error management
if ($err == $ERROR_OK)
{
	if ($nbWarnings > 0)
	{
		$status = $EXIT_STATUS_WARNING;
	}
}
else
{
	$status = $EXIT_STATUS_ERROR;
}
exit $status;



